;; Generated package description from org-gnosis.el  -*- no-byte-compile: t -*-
(define-package "org-gnosis" "0.1.1.0.20260111.154709" "Roam-like Knowledge Management System" '((emacs "27.2") (emacsql "4.0.0") (compat "29.1.4.2")) :commit "c019955b0ba2f4adcaaad0d87e7377911c1d7575" :authors '(("Thanos Apollo" . "public@thanosapollo.org")) :maintainer '("Thanos Apollo" . "public@thanosapollo.org") :keywords '("extensions") :url "https://thanosapollo.org/projects/org-gnosis/")
